import numpy as np
from .ops import Operation
from .graph import Graph
from itertools import zip_longest


def process_data(data):
  '''
    Processes the data that is stored in Tensor
    Supprted types for data - [int, float, list, np.ndarray]
    Elements in data should be float or be typecastable to float
  '''
  supported_types = [int, float, list, np.ndarray]
  if type(data) in supported_types:
    if not isinstance(data, np.ndarray):
      data = np.array(data)
    try:
      data = data.astype(float)
    except ValueError:
      raise TypeError("Elements of data should be of type float or be typecastable to float")
  else:
    raise TypeError(f"Expected data of types {supported_types} instead got {type(data)}")
  return data

def unbroadcast_data(data, orig_data_shape, broadcasted_shape):
  '''
    if data(a np object) is broadcasted during an operation, then it is unbroadcasted here
      where all dimensions where it was broadcasted are summed along that dimension to
      give the original shape of the data
  '''
  if broadcasted_shape is not None:
    dims_to_be_summed = get_dims_to_be_summed(orig_data_shape, broadcasted_shape)
    unbroadcasted_data = data.reshape(broadcasted_shape)
    for i,dim in reversed(list(enumerate(dims_to_be_summed))):
      if dim:
        unbroadcasted_data = np.sum(unbroadcasted_data, axis=i)
  else:
    unbroadcasted_data = data
  return unbroadcasted_data

def get_dims_to_be_summed(orig_data_shape, broadcasted_shape):
  '''
    True is given if it has been broadcasted along that dimension, False if not
    https://numpy.org/doc/stable/user/basics.broadcasting.html
  '''
  dims_to_be_summed = []
  zipped = zip_longest(tuple(reversed(broadcasted_shape)), tuple(reversed(orig_data_shape)), fillvalue=None)
  for dim_broadcasted, dim_orig in reversed(list(zipped)):
    if dim_broadcasted!=dim_orig:
      dims_to_be_summed.append(True)
    else:
      dims_to_be_summed.append(False)
  return dims_to_be_summed

def get_graph():
  '''
    Returns graph present in Operation.graph, if it is None, then the global graph _NG_GRAPH
      is used
  '''
  if Operation.graph is None:
    from .. import _NG_GRAPH
    graph = _NG_GRAPH
  else:
    graph = Operation.graph
  return graph


class NewGraph:
  '''
    Context Manager to create a new graph if required within an operation or
      anywhere and shouldn't interfere with the global _NG_GRAPH
  '''
  def __enter__(self):
    Operation.graph = Graph()
  
  def __exit__(self, exc_type, exc_value, exc_traceback):
    Operation.graph = None


from .utils import process_data, unbroadcast_data
from .ops import add, sub, mul, div, pow as _pow, transpose, sum as _sum, exp, dot


class Tensor:
  __slots__ = ['_data', 'requires_grad', 'grad', 'grad_fn']
  '''
    The main computational element which abstracts the np.ndarray used to
      perform all the Operation
  '''

  def __init__(self, data, requires_grad=False):
    '''
      Params:
        data:int/float/np.ndarray/[] - Data to be stored in the tensor
        requires_grad:Bool - If a Tensor requires gradient to be calculated for it or not
      
      grad:np.ndarray - The gradient for the current Tensor
      grad_fn - takes the gradient from above, uses the local gradient to give the gradient
    '''
    self.data = data
    self.requires_grad = requires_grad
    self.grad = 0. if requires_grad else None
    self.grad_fn = None
  
  def zero_grad(self):
    '''
      Resets the grad of the Tensor
    '''
    self.grad = 0. if self.requires_grad else None
  
  def backward(self, upper_grad=1., retain_graph=False):
    '''
      Starts the gradient calculation for the backward pass from the current
        Tensor
      
      Params:
        upper_grad:int/float/np.ndarray/[] - The gradient with which to start the
          gradient calculation
        retain_graph:Bool - If the graph should be retained after backward pass or flushed
    '''
    if not(self.requires_grad):
      raise ValueError("Only tensors who requires_grad can call backward")
    from .utils import get_graph
    graph = get_graph()
    upper_grad = process_data(upper_grad)
    if self.shape!=upper_grad.shape:
      raise ValueError("Shapes of grad and Tensor data must match!")
    self.grad+=upper_grad
    node = graph.get_node(self)
    node.backward(retain_graph)
    if not(retain_graph):
      graph.reset_graph() # tensors are auto-removed, this is just for redundancy / safety
  
  def _backward(self, node, retain_graph):
    '''
      The essence of autograd, final gradient calculations for the Tensor is performed here
      For each child, its gradient is taken as upper gradient, the backward function of the 
        Node of the current Tensor is executed to set the grad_fn of current Tensor
      grad_fn is executed, the grad is then unbroadcasted, if Tensor has been broadcasted
        during the Operation
      auto-removal of Tensor from the graph is performed when retain_graph is False
    '''
    from .utils import get_graph
    graph = get_graph()
    for child in node.children:
      if self.requires_grad:
        child.backward_fn(*[node.tens for node in child.parents])
        upper_grad = child.tens.grad
        if child.parent_needs_broadcasting:
          upper_grad = upper_grad.flatten()
        grad = self.grad_fn(upper_grad)
        grad = unbroadcast_data(grad, self.shape, child.parent_broadcast_shape)
        grad = grad.reshape(self.shape)
        self.grad+=grad
      if not(retain_graph) and child.are_parents_visited():
        graph.remove_tensor(child.tens)
    if not(retain_graph) and node.are_parents_visited():
      graph.remove_tensor(node.tens)
  
  def set_grad_fn(self, grad_fn):
    '''
      Sets the grad_fn for the Tensor
      if it doesn't require_grad it is set to None
    '''
    if self.requires_grad:
      self.grad_fn = grad_fn
    else:
      self.grad_fn = None

  def __add__(self, other):
    return add(self, other)
  
  def __radd__(self, other):
    return add(other, self)
  
  def __sub__(self, other):
    return sub(self, other)
  
  def __rsub__(self, other):
    return sub(other, self)
  
  def __mul__(self, other):
    return mul(self, other)
  
  def __rmul__(self, other):
    return mul(other, self)
  
  def __truediv__(self, other):
    return div(self, other)
  
  def __rtruediv__(self, other):
    return div(other, self)
  
  def __pow__(self, other):
    return _pow(self, other)
  
  def __rpow__(self, other):
    return _pow(other, self)
  
  def __pos__(self):
    return self
  
  def __neg__(self):
    return (-1*self)
  
  def dot(self, other):
    return dot(self, other)
  
  def sum(self, axis=None):
    return _sum(self, axis)
  
  def exp(self):
    return exp(self)
  
  @property
  def data(self):
    return self._data
  
  @data.setter
  def data(self, data):
    self._data = process_data(data)

  @property
  def shape(self):
    return self.data.shape
  
  @property
  def T(self):
    return transpose(self)
  
  def __getitem__(self, index):
    supported_types = [int, slice]
    if type(index) not in supported_types:
      raise TypeError(f"Expected index of {supported_types} instead got {type(index)}")
    return Tensor(self.data[index], requires_grad=self.requires_grad)
  
  def __repr__(self):
    return f'Tensor({self.data}, requires_grad={self.requires_grad})'
  
  def __str__(self):
    return f'Tensor( {self.data},\n requires_grad={self.requires_grad},\n grad_fn={self.grad_fn},\n shape={self.shape} )\n'


class Node:
  '''
    Used as an abstraction to connect the graph together and hold relationships
  '''
  __slots__ = ['tens', 'children', 'parents', 'parent_broadcast_shape', 'parent_needs_broadcasting',
              'backward_fn', 'visited']

  def __init__(self, tens):
    '''
      Params:
        tens:Tensor - The Tensor corresponding to the Node
      
      children:[Node] - List of all Nodes which uses the current Node as an operand in an Operation
      parents:[Node] - List of all Nodes(operands) that has resulted in the creation of current Node
      parent_broadcast_shape:np.shape - If the parent needs to be broadcasted from one shape to
        another, then the final broadcasted shape of the parent is stored here
      parent_needs_broadcasting:Bool - Some Operation demands that broadcasting be done, ex add, but
        some like dot must not perform broadcasting
      backward_fn:Operation.backward - Sets the grad_fn of Tensor(operand) involved in the Operation
      visited:Bool - If Node is visited or not
    '''
    self.tens = tens
    self.children = []
    self.parents = []
    self.parent_broadcast_shape = None
    self.parent_needs_broadcasting = None
    self.backward_fn = None
    self.visited = False
  
  def top_sort(self):
    '''
      Sorts the graph topologically, to perform backward pass efficiently, so that all the children's
        is calculated before the current node's gradient is calculated.
      Sorting is done by first checking if all the children are visited, if they are then the current
        node is added to sorted_tensors if not then the top_sort is performed on children
    '''
    sorted_tensors = []
    if self.are_children_visited():
      self.visited = True
      sorted_tensors.append(self.tens)
      for parent in self.parents:
        if not(parent.visited):
          sorted_tensors+=parent.top_sort()
    else:
      for child in self.children:
        if not(child.visited):
          sorted_tensors+=child.top_sort()
    return sorted_tensors
  
  def backward(self, retain_graph):
    '''
      Iterated through all sorted_tensors and performs their backward calculation

      Params:
        retain_graph:Bool - If the graph should be retained after backward pass or flushed
    '''
    from .utils import get_graph
    graph = get_graph()
    graph.reset_visited()
    sorted_tensors = self.top_sort()
    graph.reset_visited()
    for tens in sorted_tensors:
      node = graph.get_node(tens)
      node.visited = True
      tens._backward(node, retain_graph)

  def visit_all_children(self):
    '''
      All children are visited
      This will be used when Tensor.backward is calculated for a Tensor that is not the leaf TODO
    '''
    for child in self.children:
      child.visited = True

  def are_children_visited(self):
    '''
      Checks if all children are visited
    '''
    for child in self.children:
      if not(child.visited):
        return False
    return True
  
  def are_parents_visited(self):
    '''
      Checks if all parents are visited
    '''
    for parent in self.parents:
      if not(parent.visited):
        return False
    return True
  
  def add_child(self, other):
    '''
      Adds a child to the Node
    '''
    self.children.append(other)
  
  def add_parent(self, other):
    '''
      Adds a parent to the Node
    '''
    self.parents.append(other)
  
  def __repr__(self):
    return f'Node({self.tens})'
  
  def __str__(self):
    return f'Node( \n{self.tens}\nbackward_fn: {self.backward_fn}\nvisited: {self.visited}\n )'


from .layers import Layer
from ..autograd.ops import relu, sigmoid, tanh


class ReLU(Layer):
  def forward(self, inputs):
    return relu(inputs)

  def __repr__(self):
    return 'ReLU()'
  
  def __str__(self):
    return 'ReLU'


class Sigmoid(Layer):
  def forward(self, inputs):
    return sigmoid(inputs)

  def __repr__(self):
    return 'Sigmoid()'
  
  def __str__(self):
    return 'Sigmoid'


class Tanh(Layer):
  def forward(self, inputs):
    return tanh(inputs)
  
  def __repr__(self):
    return 'Tanh()'
  
  def __str__(self):
    return 'Tanh'



import numpy as np
from .node import Node


class Operation:
  '''
    Used when some input is getting transformed into an output, for functions
      where gradient calculation is required with the forward pass and the backward
      pass defined
  '''

  graph = None

  def __init__(self, operation, operand_needs_broadcasting):
    '''
      Params:
        operation:object(cls(Operation)) - Object of an operation that inherits Operation
        operand_needs_broadcasting:Bool - If the operand_needs_broadcasting for a pareticular
          Operation
    '''
    self.operation = operation
    self.operand_needs_broadcasting = operand_needs_broadcasting
  
  def process_operands(self, operands):
    '''
      All operands are converted to Tensor

      Params:
        operands:(any class that is supported/Tensor)
    '''
    from .tensor import Tensor
    operands = list(operands)
    for i,operand in enumerate(operands):
      if not isinstance(operand, Tensor):
        operands[i] = Tensor(operand)
    return tuple(operands)
  
  def get_tensors(self, *operands):
    '''
      Returns the processed operands as tuple of Tensors

      Params:
        operands:*args(any class that is supported/Tensor)
    '''
    tensors = self.process_operands(operands)
    if len(tensors)==0:
      return None
    elif len(tensors)==1:
      return tensors[0]
    else:
      return tensors
  
  def get_broadcast_shape(self, *tensors):
    '''
      If the tensors can be broadcasted and operand_needs_broadcasting, then the broadcasted
        shape is returned, else None
      
      Params:
        tensors:*args(Tensor)
    '''
    if self.operand_needs_broadcasting:
      try:
        return np.broadcast_shapes(*(tens.data.shape for tens in tensors))
      except ValueError:
        return None
    else:
      return None
  
  def result_requires_grad(self, tensors):
    '''
      Checks if the result requires_grad given the operands of the Operation, if atleast
        one operand requires_grad, then result will also have requires_grad
      
      Params:
        tensors:(Tensor)
    '''
    for tens in tensors:
      if tens.requires_grad:
        return True
    return False
  
  def get_result_tensor(self, result, *tensors):
    '''
      Returns the result tensor of the Operation
      Creates a Node for the result_tensor with parent_broadcast_shape and parent_needs_broadcasting
      Adds the edges to the graph

      Params:
        result:np object - Result after performing a raw numpy operation
        tensors:*args(Tensor)
    '''
    from .tensor import Tensor
    from .utils import get_graph
    graph = get_graph()
    result = result.astype(np.ndarray)
    result_tensor = Tensor(result, self.result_requires_grad(tensors))
    result_node = Node(result_tensor)
    result_node.parent_needs_broadcasting = self.operand_needs_broadcasting
    result_node.backward_fn = self.operation.backward
    result_node.parent_broadcast_shape = self.get_broadcast_shape(*tensors)
    graph.add_edge(result_node, tensors)
    return result_tensor


# <------------ADD------------>

class Add(Operation):
  def __init__(self):
    super().__init__(self, True)

  def forward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(tens1.data+tens2.data, tens1, tens2)

  def backward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    tens1.set_grad_fn(lambda ug:ug)
    tens2.set_grad_fn(lambda ug:ug)

def add(tens1, tens2):
  return Add().forward(tens1, tens2)


# <------------SUB------------>

class Sub(Operation):
  def __init__(self):
    super().__init__(self, True)
  
  def forward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(tens1.data-tens2.data, tens1, tens2)
  
  def backward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    tens1.set_grad_fn(lambda ug:ug)
    tens2.set_grad_fn(lambda ug:-ug)

def sub(tens1, tens2):
  return Sub().forward(tens1, tens2)


# <------------MUL------------>

class Mul(Operation):
  def __init__(self):
    super().__init__(self, True)
  
  def forward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(tens1.data*tens2.data, tens1, tens2)
  
  def backward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    broadcast_shape = self.get_broadcast_shape(tens1, tens2)
    tens1.set_grad_fn(lambda ug:np.dot(np.diag(np.ndarray.flatten(np.broadcast_to(tens2.data, broadcast_shape))), ug))
    tens2.set_grad_fn(lambda ug:np.dot(np.diag(np.ndarray.flatten(np.broadcast_to(tens1.data, broadcast_shape))), ug))

def mul(tens1, tens2):
  return Mul().forward(tens1, tens2)


# <------------DIV------------>

class Div(Operation):
  def __init__(self):
    super().__init__(self, True)
  
  def forward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(tens1.data/tens2.data, tens1, tens2)
  
  def backward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    broadcast_shape = self.get_broadcast_shape(tens1, tens2)
    tens1.set_grad_fn(lambda ug:np.dot(np.diag(np.ndarray.flatten(np.broadcast_to(1/tens2.data, broadcast_shape))), ug))
    tens2.set_grad_fn(lambda ug:np.dot(np.diag(np.ndarray.flatten(np.broadcast_to((-1*tens1.data)/np.power(tens2.data, 2), broadcast_shape))), ug))

def div(tens1, tens2):
  return Div().forward(tens1, tens2)


# <------------DOT------------>

class Dot(Operation):
  def __init__(self):
    super().__init__(self, False)
  
  def forward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(np.dot(tens1.data, tens2.data), tens1, tens2)
  
  def backward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    tens1.set_grad_fn(lambda ug:np.dot(ug, tens2.data.T))
    tens2.set_grad_fn(lambda ug:np.dot(tens1.data.T, ug))

def dot(tens1, tens2):
  return Dot().forward(tens1, tens2)


# <------------EXP------------>

class Exp(Operation):
  def __init__(self):
    super().__init__(self, False)
  
  def forward(self, tens):
    tens = self.get_tensors(tens)
    return self.get_result_tensor(np.exp(tens.data), tens)
  
  def backward(self, tens):
    tens = self.get_tensors(tens)
    tens.set_grad_fn(lambda ug:np.exp(tens.data)*ug)

def exp(tens):
  return Exp().forward(tens)


# <------------LOG------------>

class Log(Operation):
  def __init__(self):
    super().__init__(self, False)
  
  def forward(self, tens):
    tens = self.get_tensors(tens)
    return self.get_result_tensor(np.log(tens.data), tens)
  
  def backward(self, tens):
    tens = self.get_tensors(tens)
    tens.set_grad_fn(lambda ug:(1/tens.data)*ug)

def log(tens):
  return Log().forward(tens)


# <------------POW------------>

class Pow(Operation):
  def __init__(self):
    super().__init__(self, True)
  
  def forward(self, tens1, tens2):
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(np.power(tens1.data, tens2.data), tens1, tens2)
  
  def backward(self, tens1, tens2):
    result = np.power(tens1.data, tens2.data)
    tens1, tens2 = self.get_tensors(tens1, tens2)
    tens1.set_grad_fn(lambda ug:(np.power(tens1.data, tens2.data-1) * tens2.data).flatten()*ug)
    tens2.set_grad_fn(lambda ug:(result*np.log(tens1.data)).flatten()*ug)

def pow(tens1, tens2):
  return Pow().forward(tens1, tens2)


# <------------SUM------------>

class Sum(Operation):
  def __init__(self, axis=None):
    super().__init__(self, True)
    self.axis = axis
  
  def forward(self, tens):
    tens = self.get_tensors(tens)
    return self.get_result_tensor(np.sum(tens.data, axis=self.axis), tens)
  
  def backward(self, tens):
    tens = self.get_tensors(tens)
    
    def grad_backward(ug):
      tens_shape = list(tens.shape)
      if self.axis is not None:
        try:
          tens_shape[self.axis] = 1
        except IndexError:
          pass
        lg = 1
      else:
        lg = np.ones(tens.shape)

      if self.axis is not None:
        grads = np.dot(lg, ug)
        try:
          num_repeat = tens.shape[self.axis]
        except IndexError:
          num_repeat = 1
        grads = grads[np.newaxis]
        grads = np.concatenate([grads]*num_repeat)
      else:
        grads = lg*ug
      return grads
    tens.set_grad_fn(grad_backward)

def sum(tens, axis=None):
  return Sum(axis).forward(tens)


# <------------TRANSPOSE------------>

class Transpose(Operation):
  def __init__(self):
    super().__init__(self, False)
  
  def forward(self, tens):
    tens = self.get_tensors(tens)
    return self.get_result_tensor(tens.data.T, tens)

  def backward(self, tens):
    tens = self.get_tensors(tens)
    tens.set_grad_fn(lambda ug:ug.T)

def transpose(tens):
  return Transpose().forward(tens)


# <------------RELU------------>

class ReLU(Operation):
  def __init__(self):
    super().__init__(self, False)
  
  def forward(self, tens):
    tens = self.get_tensors(tens)
    return self.get_result_tensor(np.maximum(0, tens.data), tens)
  
  def backward(self, tens):
    tens = self.get_tensors(tens)
    tens.set_grad_fn(lambda ug:np.where(tens.data>=0, 1, 0)*ug)

def relu(tens):
  return ReLU().forward(tens)


# <------------SIGMOID------------>

class Sigmoid(Operation):
  def __init__(self):
    super().__init__(self, False)
  
  def forward(self, tens):
    tens = self.get_tensors(tens)
    return self.get_result_tensor(1/(1+np.exp(-tens.data)), tens)
  
  def backward(self, tens):
    tens = self.get_tensors(tens)
    result = 1/(1+np.exp(-tens.data))
    tens.set_grad_fn(lambda ug:(result*(1-result))*ug)

def sigmoid(tens):
  return Sigmoid().forward(tens)


# <------------TANH------------>

class Tanh(Operation):
  def __init__(self):
    super().__init__(self, False)
  
  def forward(self, tens):
    tens = self.get_tensors(tens)
    return self.get_result_tensor(np.tanh(tens.data), tens)
  
  def backward(self, tens):
    tens = self.get_tensors(tens)
    result = np.tanh(tens.data)
    tens.set_grad_fn(lambda ug:(1-np.power(result,2))*ug)

def tanh(tens):
  return Tanh().forward(tens)


def get_batches(inputs, targets, num_examples, batch_size=None):
  '''
    Split the inputs and their corresponding targets into batches for efficient
      training
    TODO: Currently only works for 1D data, have to fix it to accomodate all
      dimensional data
  '''
  if batch_size is not None:
    if batch_size > num_examples:
      raise ValueError("Batch size cannot be greater than the number of examples")
    elif batch_size<0:
      raise ValueError("Batch size cannot be negative")
    elif batch_size==0:
      raise ValueError("Batch size cannot be zero")
  else:
    batch_size = num_examples
  start = 0
  for end in range(batch_size, num_examples, batch_size):
    yield (inputs[start:end], targets[start:end])
    start = end
  if end!=num_examples-1:
    yield (inputs[end:], targets[end:])


from .autograd import tensor, NewGraph
from .autograd import add, sub, mul, div, pow, exp, log, dot, sum, transpose
from .autograd.graph import Graph


global _NG_GRAPH
_NG_GRAPH = Graph()
'''
  _NG_GRAPH is the global graph object used to construct backprop graphs
'''


from ..autograd import tensor, dot
import numpy as np


class Container:
  __slots__ = ['layers']
  '''
    This acts as a container for Layer/s
  '''

  def __call__(self, inputs):
    return self.forward(inputs)

  def get_params(self):
    '''
      Goes through all the layers in the Container and gets the params of each Layer
    '''
    params = []
    for layer in self.layers:
      params+=layer.get_params()
    return params
  
  def __repr__(self):
    layers = []
    for layer in self.layers:
      layers.append(f'{layer.__str__()}')
    layers_repr = ', '.join(layers)
    return layers_repr
  
  def __str__(self):
    layers = []
    for layer in self.layers:
      layers.append(f'{layer.__repr__()}')
    layers_str = ', '.join(layers)
    return layers_str


class Layer:
  __slots__ = []
  '''
    Performs some kind of computation taking in inputs and giving out outputs
  '''

  def __call__(self, inputs):
    return self.forward(inputs)

  def get_params(self):
    '''
      If any of the attributes in a Layer is instance of Param, then it is automatically
        considered as a param for the whole model
    '''
    params = []
    for attr in dir(self):
      attr_val = self.__getattribute__(attr)
      if isinstance(attr_val, Param):
        params.append(attr_val)
    return params


class Param(tensor):
  __slots__ = []
  '''
    Just an alias for Tensor, so that when params are gathered for a Layer, only these
      are automatically considered for param, while ignoring some helper Tensors which aren't
      necessarily param
  '''

  def __init__(self, data, requires_grad=False):
    super().__init__(data, requires_grad)
  
  def __str__(self):
    return f'Param({super().__str__()})'
  
  def __repr__(self):
    return f'Param({super().__repr__()})'


class Sequential(Container):
  __slots__ = ['layers']
  '''
    Outputs of one layer are passed as inputs to the next layer, sequentially
  '''
  
  def __init__(self, *args):
    self.layers = args
  
  def forward(self, inputs):
    for layer in self.layers:
      output = layer(inputs)
      inputs = output
    return output
  
  def __str__(self):
    return f'Sequential(\n{super().__str__()}\n)'
  
  def __repr__(self):
    return f'Sequential(\n{super().__repr__()}\n)'


class Linear(Layer):
  __slots__ = ['num_in', 'num_out', 'weights', 'bias']
  '''
    Implements a fully connected layer
  '''

  def __init__(self, num_in, num_out):
    self.num_in = num_in
    self.num_out = num_out
    self.weights = Param(np.random.randn(num_out, num_in), requires_grad=True)
    self.bias = Param(np.zeros((num_out, 1)), requires_grad=True)
  
  def forward(self, inputs):
    return dot(self.weights, inputs) + self.bias
  
  def __repr__(self):
    return f'Linear({self.num_in}, {self.num_out})'
  
  def __str__(self):
    return f'Linear in:{self.num_in} out:{self.num_out}'


from .layers import Container, Layer


class Model:
  __slots__ = ['model', 'layer_types']

  def __init__(self, model):
    self.model = model
    self.layer_types = (Container, Layer)
  
  def __call__(self, inputs):
    return self.forward(inputs)

  def get_layers(self):
    '''
      In the attributes of the model object, it searches if any of them are
        subclasses of Container/Layer, if so it considers it as a layer
    '''
    layers = []
    for attr in dir(self.model):
      model_attr_val = self.model.__getattribute__(attr) 
      if issubclass(model_attr_val.__class__, self.layer_types):
        layers.append(model_attr_val)
    return layers
  
  def get_params(self):
    '''
      Gathers the params of the whole model by iterating through all layers and getting their params
    '''
    params = []
    layers = self.get_layers()
    for layer in layers:
      params+=layer.get_params()
    return params
  
  def __repr__(self):
    return f'Model( {[str(layer) for layer in self.layers]} )'
  
  def __str__(self):
    return f'Model( {[str(layer) for layer in self.layers]} )'


from .node import Node


class Graph:
  '''
    The graph that is constructed during the forward pass, and used by the backward
      pass to calculate gradients through automatic differentiation
  '''
  __slots__ = ['nodes_dict']
  
  def __init__(self):
    '''
      nodes_dict contains a Tensor as the key and its respective Node as the value
    '''
    self.nodes_dict = {}
  
  def add_edge(self, result_node, operands):
    '''
      Adds edges between the result_node, which is created during an Operation, and the
      operands that produced the result.

      Params:
        result_node:Node - node that is created in Operation.get_result_tensor
        operands:[Tensor] - All the operands for an Operation
    '''
    self.add_node(result_node)
    for operand in operands:
      if self.get_node(operand) is None:
        self.add_tensor(operand)
      operand_node = self.get_node(operand)
      result_node.add_parent(operand_node)
      operand_node.add_child(result_node)
  
  def add_node(self, node):
    '''
      Adds a Node to the graph

      Params:
        node:Node
    '''
    self.nodes_dict[node.tens] = node

  def get_node(self, tens):
    '''
      Returns the Node corresponding to the Tensor

      Params:
        tens:Tensor
    '''
    return self.nodes_dict.get(tens)
  
  def add_tensor(self, tens):
    '''
      Adds a Tensor to the graph. A new node is created for the Tensor

      Params:
        tens:Tensor
    '''
    self.nodes_dict[tens] = Node(tens)
  
  def remove_tensor(self, tens):
    '''
      Removes a Tensor from the graph

      Params:
        tens:Tensor
    '''
    self.nodes_dict.pop(tens)
  
  def reset_visited(self):
    '''
      Resets Node.visited for each Node in the graph
    '''
    for node in self.nodes_dict.values():
      node.visited = False
  
  def reset_graph(self):
    '''
      Resets the whole graph clears the nodes_dict
    '''
    self.nodes_dict = {}
  
  def __repr__(self):
    return 'Graph()'
  
  def __str__(self):
    return f'Graph( {self.nodes_dict} )'


from ..autograd import sum, log


class Loss:
  def __call__(self, outputs, targets):
    return self.forward(outputs, targets)


class MSE(Loss):
  '''
    Mean Squared Error
  '''
  def forward(self, outputs, targets):
    num_examples = outputs.shape[-1]
    cost = (1/(2*num_examples))*sum((outputs-targets)**2)
    return cost
  
  def __repr__(self):
    return f'MSE()'
  
  def __str__(self):
    return 'MeanSquaredError'


class BCE(Loss):
  '''
    Binary Cross Entropy

    epsilon used here to prevent log0
  '''
  def forward(self, outputs, targets):
    epsilon = 1e-5
    num_examples = outputs.shape[-1]
    entropy = ((outputs*log(targets+epsilon)) + ((1-outputs)*(log(1-targets+epsilon))))
    cost = (-1/num_examples)*(sum(entropy))
    return cost
  
  def __repr__(self):
    return f'BCE()'
  
  def __str__(self):
    return 'BinaryCrossEntropy'


class Optimizer:
  def zero_grad(self):
    '''
      Since after loss.backward, only Tensors in memory are the params, only their
        gradients are reset since everytime a new graph is dynamically created
    '''
    for param in self.params:
      param.zero_grad()


class GD(Optimizer):
  '''
    Vanilla Gradient Descent
  '''
  def __init__(self, params, lr):
    self.params = params
    self.lr = lr
  
  def step(self):
    for param in self.params:
      if param.requires_grad:
        param.data -= (self.lr*param.grad)
  
  def __repr__(self):
    return f'GD(params={self.params}, lr={self.lr})'
  
  def __str__(self):
    return f'GD(params={self.params}, lr={self.lr})'